// Time-stamp: <05/12/01 11:55:57 nijw>
/** @file        */
#ifndef RTREE__H_IJW
#define RTREE__H_IJW

#include <list>
// utilities
#include "utilityfunctionals.h" // for DeleteObject
// Genetics Headers
#include "gentldeclr.h"
#include "recombinationrates.h"
#include "rnodet.h"
#include "nodet.h"
#include "coalescent.h"

namespace GenTL {
  /** The default collector class - just counts the recombinations  *
   * Other collector classes can easily be written - they just have 
   * have the operator()(time, event,which) method which is used within 
   * rtree::simulate                                                    */
  struct recombinations {
    /** default constructor                                             */
    recombinations():rec_(0){};
    /** function overload operator - needed for any COLLECTOR structure */
    void operator()(double time, EventType e,int which=0) {
      if (e==Recombination) rec_ +=1;
    };
    /** function operator that just returns the result                  */
    long operator()() {
      return result();
    };
    /** Return the number of recombinations                             */
    long result() {
      return rec_;
    };
    /** Data -- keeps the number of recombinations                      */
    long rec_;
  };
  struct recombinationLocations {
    /** default constructor                                             */
    recombinationLocations() {};
    /** function overload operator - needed for any COLLECTOR structure */
    void operator()(double time, EventType e,int which=0) {
      if (e==Recombination) where_.push_back(which);
    };
    /** function operator that just returns the result                  */
    std::list<int> &operator()() {
      return where_;
    };
    /** Return the list of recombination locations                      */
    std::list<int> &result() {
      return where_;
    };
    /** return count */
    int count() { 
      return where_.size();
    }
    /** Data -- keeps the number of recombinations                      */
    std::list<int> where_;
  };
 

  /** ARG (ancestral recombination graph) class.  The general class for 
      simulating recombination trees                                    */
  template<class T,typename COLLECTOR> 
  class ARG {
  public:
    /** constructor which creates its own coalescent object             */
    ARG(int ss, int nl,const std::vector<double> &recomb_rates,growthmodel *g,rng &r,COLLECTOR &clt)
      :sample(ss,nl)
      ,root(nl)
      ,collect(clt) {
      coalescent cc(ss,g,r);
      simulate(cc,recomb_rates,r);
    };
    /** Constructor which requires a coalescent object (which includes the sample size          */
    ARG(int nl,const std::vector<double> &recomb_rates, coalescent &c,rng &r,COLLECTOR &clt)
      :root(nl), sample(c.nleft(),nl) ,collect(clt) {
      simulate(c,recomb_rates,r);
    };
    /** dtor                 
     * This particular formulation is that of Meyers - effective STL                             */
    ~ARG() {
      for_each(ancestors.begin(),ancestors.end(),
	       DeleteObject());
    }
    /** prune the ancestral recombination graph  */
    void prune(const std::vector<size_t> &wh) {
     std::list<recombnode<T> *> used;
     for (size_t i=0;i<wh.size();i++) {
       assert(wh[i]>=0&&wh[i]<sample.size());
       used.insert(used.end(),sample[wh[i]].l.begin(),sample[wh[i]].l.end());
     }
     typename std::list<rnode<T>* >::iterator ii=ancestors.begin();
     while (ii!=ancestors.end()) {
       //    (*ii)->l.print_active(std::cerr,"");

       typename recomblist<T>::itor  
	 left=std::find(used.begin(),used.end(),(*ii)->left_),
	 right=std::find(used.begin(),used.end(),(*ii)->right_);
       if (left!=used.end()) {
	 if (right!=used.end()) {
	   // both present 
       	   (*ii)->l.fixlist(*left,*right); 
	   used.insert(left,(*ii)->l.begin(),(*ii)->l.end());
	   used.erase(left);
	   used.erase(right);
	   ii++;
	 } else {
	   // left present but not right
	   rnode<T> *newd=(*ii)->left_->coal;
	   (*ii)->l.fixlist(*left); // remove all those sections that are not needed 
	   used.insert(used.end(),(*ii)->l.begin(),(*ii)->l.end());
	   used.erase(left);
	   
	   typename recomblist<T>::itor iif=find(newd->l.begin(),newd->l.end(),(*ii)->left_);
	   //   position recombpos=(*ii)->left_.recombination;

	   assert (iif!=newd->l.end());
	   newd->l.splice(iif,(*ii)->l);
	   newd->l.erase(iif);
	   delete(*ii);
	   ii=ancestors.erase(ii);
	 }
       } else if (right!=used.end()) {
	 // right present but not left

	 rnode<T> *newd=(*ii)->right_->coal;
	 
	 (*ii)->l.fixlist(*right);  // remove all the recombnodes that are not needed
	 used.insert(used.end(),(*ii)->l.begin(),(*ii)->l.end());
	 used.erase(right);

	 typename recomblist<T>::itor iif=find(newd->l.begin(),newd->l.end(),(*ii)->right_);
	 //	 position recombpos=(*ii)->right_.recombination;
	 assert (iif!=newd->l.end());
	 newd->l.splice(iif,(*ii)->l);
	 newd->l.erase(iif);
	 
	 delete(*ii);
	 ii=ancestors.erase(ii);
       } else {
	 // neither present
	 delete(*ii);
	 ii=ancestors.erase(ii);
       }
     }
     remakeroots(wh);
     checkroots(wh);
     ii=ancestors.begin();
//      while (ii!=ancestors.end()) {
//        (*ii)->l.print_active(std::cerr,"");
//        ii++;
//      }
    }
    /** remake the roots after a pruning operation                                            */
    void remakeroots(const std::vector<size_t> &wh) {
      for (size_t i=0;i<root.size();i++) {
	typename std::list<rnode<T>* >::reverse_iterator ri=ancestors.rbegin();
	while (ri!=ancestors.rend()) {
	  if ((*ri)->left_->IsActive(i)&&(*ri)->right_->IsActive(i)) {
	    root[i]=*ri;
	    break;
	  }
	  ri++;
	}
      }
    }
    TNT::Array2D<position> sharedsections(const std::vector<int> &wh, position pos)
    {
      TNT::Array2D<position> minmax(wh.size(),2);
      // extract the tree at a position
      node<rnode<T> *> *myroot=extract_node(root[pos],pos,&sample[0]);
            
      for (size_t i=0;i<wh.size();i++) {
	// find the node at the correct depth for wh (wasteful...)
	
	node<rnode<T> *> *current=myroot->find_node(wh[i]);
	//	std::cerr << current->label << " " << current->data() << std::endl;
	if (current==0) {
	  std::cerr << "label " << wh[i] <<" not found";
	    exit(EXIT_FAILURE);
	}
	node<rnode<T> *> *anc=current->up;
	//	std::cerr << anc->label << " " << anc->data() << std::endl;
	
	position lf=anc->data()->left_->active.first();
	position lm=anc->data()->left_->active.last();
	position rf=anc->data()->right_->active.first();
	position rm=anc->data()->right_->active.last();
	
	minmax[i][0]=std::max(lf,rf);
	minmax[i][1]=std::min(lm,rm);
	
      }
      recursively_destroy_node(myroot);
      return minmax;
    }
    /** What is the time of the root at position loc                                             */
    double roottime(position loc) {
      return root[loc]->time;
    };
    /** What is the length of the tree at position loc                                           */
    double length(position loc) {
      return root[loc]->length(loc);
    } 
    /** How many coalescent nodes are there in this tree                                         */
    long CoalescentNodes() const { 
      return ancestors.size();
    };
    /** is this node a root?                                              */
    bool isroot(rnode<T> *any) {
      if (find(root.begin(),root.end(),any)!=root.end()) return true;
      return false;
    }
    /** root for which poositions?                                        */
    std::vector<int> rootwhichpositions(rnode<T> *n) {
      std::vector<int> pos;
      for (size_t i=0;i<root.size();i++) {
	if (root[i]==n) pos.push_back(i);
      }
      return pos;
    }
    /** randomise the node ordering                                                              */
    void randomise(rng &ran) {
      typename std::list<rnode<T> *>::iterator i=ancestors.begin();
      while (i != ancestors.end()) {
	if (ran()<0.5) {
	  recombnode<T> *tmp=(*i)->left_;
	  (*i)->left_=(*i)->right_;
	  (*i)->right_=tmp;
	} 
	i++;
      }
    }
    /** check that the roots have descendents to the left and right at each position             */
    bool checkroots(const std::vector<size_t> &wh) {
      for (size_t i=0;i<root.size();i++) {
	assert((root[i]->left_->IsActive(i)&&root[i]->right_->IsActive(i)));
	bool left=root[i]->left_->coal->present(wh,&sample[0]);
	bool right=root[i]->right_->coal->present(wh,&sample[0]);
	if (!(left&&right)) {
	  std::cerr << " problem with root of site " << i << std::endl;
	  std::cerr << left << "-" << right  << std::endl; 
	  exit(0);
	}
      }
      return true;
    }
    /** Simulate the tree                                                                       */
    void simulate(coalescent &c, const std::vector<double> recomb_rates,rng &r);


    std::vector<rnode<T>* > root;    //< a vector of roots
    std::vector<rnode<T> > sample;   //< the sampled leaves
    std::list<rnode<T>* > ancestors; //< ancestral coalescences
    /** How many loci?                                                                          */
    int nloc() const  {
      return sample[0].last()->active.last()+1;
    }
    int sites() const  {
      return root.size();
    }
  private:
    COLLECTOR &collect;
    ARG();
  };

  /** A utility function to do the simulations for all constructors                             */
  template <class T,typename COLLECTOR>
  void ARG<T,COLLECTOR>::simulate(coalescent &c, const std::vector<double> recomb_rates,rng &r)
  {
#ifdef CHECK
    std::cerr << "Before simulation have " << c.nleft() << " lines" << std::endl;
#endif
    double t=0.0;
    lines_of_descent<T> lines(sample);
    rec_rates<T> rr(c.nleft(),recomb_rates.size()+1,recomb_rates,r);
    for (;;) {
#ifdef CHECK
      std::cerr << "================================================" << std::endl 
		<< "Time is now " << t << std::endl;
      lines.print(std::cout);
      std::cerr << "total rate "<<rr.rate()<<std::endl;
      rr.print_rates(std::cerr);
      rr.check(lines);
      c.print(std::cerr);
#endif    
      double NextRecombination=rr.next(t);
      double NextEvent=c.next(t);
#ifdef CHECK
      std::cerr << "NextRecombination = " << NextRecombination 
		<< "Next Event " << NextEvent << std::endl;
#endif
      if (NextEvent<NextRecombination) {
	t += NextEvent;
	std::pair<int,int> chromo = c.coalchromo();
#ifdef CHECK
	std::cout << "coalescing "<< chromo.first 
		  << " and " << chromo.second << std::endl;
#endif
	if (chromo.first!=chromo.second) { // a coalescence 
	  collect(t,Coalescence);
	  ancestors.push_back(new rnode<T>(chromo,t,rr,root,lines));
	  // Now fix lines 
	  int removed=lines.coal(ancestors.back(),chromo);
	  c.UpdateCoalescence(chromo,removed);    
	  rr.update_coal(chromo,lines[chromo.first]->active,removed); 
	} else collect(t,Migration);  // other
      } else {//	  recombination
	t += NextRecombination;
	std::pair <int,int> wh=rr.recombination(lines);
	collect(t,Recombination,wh.second);
#ifdef CHECK
	std::cout << "recombination for line " << wh.first
		  << " at position " << wh.second << std::endl;
#endif
	recombnode <T> *nw=
	  lines[wh.first]->coal->recombination(lines[wh.first],wh.second,t);	    
      
	lines.push_back(nw);
	c.UpdateRecombination(wh.first);
	rr.update_recombination(wh.first,lines);
      }
      if (c.nleft()<=1) return;
    }
  }
  /** recursive function to extract the underlying tree at a given position */
  template<typename T>
    node<rnode<T> *> *extract_node(rnode<T> *here,int position, rnode <T> *first)
    {
      static int label =1000000;
      if (here->isleaf()) 
	return new node<rnode<T> *>(here-first,here);
      else 
	return new  node<rnode<T> *>(here->time
				,extract_node(here->left(position),position,first)
				     ,extract_node(here->right(position),position,first)
				     ,++label,here);
    }

 

}
#endif   

