/** @file */
#ifndef GROWTHMODEL_H_
#define GROWTHMODEL_H_

#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include <utility>     // for pair
//#include "utilcpp/gsl_rand.H"
#include "gsl_rand.H"

namespace GenTL {
  /** abstract base class                                                */
  class growthmodel {
  public:
    virtual double nextcoal(const int n,const double t,rng &r)=0;
    virtual double nextcoalrel(const int n,const double t,const double rel, rng &r)=0;
    /** what is the next coalescence for multiple populations */
    std::pair<double,int> nextcoalv(const std::vector<int> &n,double t,rng &r);
    /** what is the next coalescence for multiple populations with different population sizes */
    std::pair<double,int> nextcoalvrel(const std::vector<int> &n, const std::vector<double> rel
				       ,const double t,rng &r);
    virtual ~growthmodel(){};
    virtual growthmodel *clone()=0;
    virtual void print(std::ostream &o) const=0; 
  };
  /** constant population size                                           */
  class constant_size: public growthmodel {
  public:
    constant_size(){};
    ~constant_size(){};
    constant_size *clone()  {
      constant_size *gm=new constant_size();
      return gm;
    }	    
    double nextcoal(const int n,const double t,rng &r){
      return r.sexp()*2.0/(double(n*(n-1)));
    }
    double nextcoalrel(const int n,const double t,const double rel,rng &r){
      return rel*r.sexp()*2.0/(double(n*(n-1)));
    }
    void print(std::ostream &o) const {
      o << "constant" << std::endl; // with relative size "<<relsize_<<std::endl;
    };
  }; 
  /** class to generate coalescence times for an exponentially growing population  */
  class exponentialgrowth: public growthmodel {
  public:
    exponentialgrowth(double gr):
      growthrate_(gr) {
    };
    exponentialgrowth(const std::vector<double> &x)
      :growthrate_(x[0]){
    };
    
    ~exponentialgrowth(){};
    exponentialgrowth *clone()  {
      exponentialgrowth *gm=new exponentialgrowth(growthrate_);
      return gm;
    }	
    double nextcoal(const int n,const double last,rng &r){
      if (fabs(growthrate_)>1E-3) {
	return log(
		   exp(growthrate_*last)-log(r())*2.*growthrate_/((double)n*((double)n-1.0)) 
		   )/growthrate_-last;
      } else return r.sexp()*2.0/(double(n*(n-1)));
    }
    double nextcoalrel(const int n,const double last,const double rel, rng &r){
      if (fabs(growthrate_)>1E-3) {
	return log(
		   exp(growthrate_*last)-log(r())*2.*growthrate_*rel/((double)n*((double)n-1.0)) 
		   )/growthrate_-last;
      } else return rel*r.sexp()*2.0/(double(n*(n-1)));
    }

    void print(std::ostream &o) const {
      o << "exponential("<<growthrate_<<")" 
	<< std::endl;
    };
  private:
    double growthrate_;
  };
  /** exponentially growing from a base                                                */
  class expfrombase: public growthmodel {
  public:
    expfrombase(double gr,double t0)
      :t_0_(t0),growthrate_(gr) {
    };
      expfrombase(const std::vector<double> &x)
	:t_0_(x[1]),growthrate_(x[0]) {
    };
    ~expfrombase(){};
    expfrombase *clone()  {
      expfrombase *gm=new expfrombase(growthrate_,t_0_);
      return gm;
    }	
    double nextcoal(const int n,const double last,rng &r) {
      if (last>=t_0_) {
	return -log(r())*2.0/(double(n)*double(n-1));
      } else {
	double t = log(
			   exp(growthrate_*last)-log(r()*exp(growthrate_*t_0_)*2.*growthrate_
						     /(double(n)*double(n-1.0))))/growthrate_-last;
	if (t+last> t_0_) 
	  return (t_0_-last) -log(r())*2.0/(double(n)*double(n-1));
	else return t;
      }
    }	
    double nextcoalrel(const int n,const double last,const double rel,rng &r) {
      if (last>=t_0_) {
	return -rel*log(r())*2.0*rel/(double(n)*double(n-1));
      } else {
	double t = log(
		       exp(growthrate_*last)-log(r()*exp(growthrate_*t_0_)*2.*growthrate_*rel
						 /(double(n)*double(n-1.0))))/growthrate_-last;
	if (t+last> t_0_) 
	  return (t_0_-last) -log(r())*2.0*rel/(double(n)*double(n-1));
	else return t;
      }
    }		
    void print(std::ostream &o) const {
      o << "expfrombase(" 
	<<growthrate_ << "," << t_0_<<")"<<std::endl; 
    };
  private:
    double t_0_,growthrate_;
  };
  /** a simple three parameter bottleneck (t_0,a,t_1)                        */
  /*  Size 1 until time t_0 in the past, then changes to size a, for a further
   * time t_1, before returning to size 1                                   
   *       
   *
   *           |                                    |  
   *           |                                    |  
   *           |                                    |  
   *                            |   |
   *                            |   |    t1
   *                            | a |
   *           |                                    |  
   *           |                                    |  
   *           |                                    |  
   *    t_0    |                                    |  
   *           |                                    |  
   *           |                                    |                          */
  class bottleneck: public growthmodel {
  public:
    bottleneck(double t0,double a,double t1)
      :t_0_(t0),t_1_(t1),a_(a) {
    };
      bottleneck(const std::vector<double> &x)
      :t_0_(x[0]),t_1_(x[2]),a_(x[1]) {
    };
	~bottleneck(){};
    growthmodel *clone()  {
      growthmodel *gm=new bottleneck(t_0_,a_,t_1_);
      return gm;
    }	
    double nextcoal(const int n,const double last,rng &r) {
      if (last >= t_0_+t_1_) {                             // back to the original size
	return -log(r())*2.0/(double(n)*double(n-1));
      } else if (last >= t_0_) {                            // in the bottleneck
	double t= -a_*log(r())*2.0/(double(n)*double(n-1));
	if (t>t_1_) return t_0_+t_1_-last+nextcoal(n,t_1_,r);
	else return t;
      } else {                                             // before the bottleneck
	double t= -log(r())*2.0/(double(n)*double(n-1));
	if (last+t>t_0_) return t_0_-last+ nextcoal(n,t_0_,r);
	else return t;
      }
    }
      double nextcoalrel(const int n,const double last,const double rel,rng &r) {
	if (last >= t_0_+t_1_) {                             // back to the original size
	  return -log(r())*2.0*rel/(double(n)*double(n-1));
	} else if (last >= t_0_) {                            // in the bottleneck
	  double t= -a_*log(r())*2.0*rel/(double(n)*double(n-1));
	  if (t>t_1_) return t_0_+t_1_-last+nextcoalrel(n,t_1_,rel,r);
	  else return t;
	} else {                                             // before the bottleneck
	  double t= -log(r())*2.0*rel/(double(n)*double(n-1));
	  if (last+t>t_0_) return t_0_-last+ nextcoalrel(n,t_0_,rel,r);
	  else return t;
	}
      }	
  
      void print(std::ostream &o) const {
	o << "bottleneck(" 
	  << t_0_ << "," << a_<<"," << t_1_ <<")" << std::endl; 
      };
    private:
      double t_0_,t_1_,a_;
    };
  /** A general population size change model */
  /* starts with size 1, then changes to a_0 at time t_0, then changes to size a_1 after
   * a further time t_1 and so on ... with size a_k from sum_0^k t_i                        */
  class piecewise: public growthmodel {
  public:
    piecewise(const std::vector<double> &x, double rel=1.0)
       {
      std::vector<double>::const_iterator i=x.begin();
      while (i!=x.end()) {
	t.push_back(*i++);
	a.push_back(*i++);
      }
    }
      ~piecewise(){};
      piecewise *clone()  {
	piecewise *gm=new piecewise(*this);
	return gm;
      }	
      double nextcoal(const int n,const double last,rng &r) {
	return 1.0;
      }
      double nextcoalrel(const int n,const double last,const double rel,rng &r) {
	return 1.0;
      }
      void print(std::ostream &o) const {
	o << "piecewise(";
	for (unsigned int i=0;i<t.size()-1;i++) {
	  o << t[i] << "," << a[i] <<",";
	}
	o << t[t.size()-1] <<"," << a[t.size()-1] <<")";
      };
  private:
      std::vector<double> t;
      std::vector<double> a;
    };
}
GenTL::growthmodel *gmread(std::string &s);
std::ostream &operator<<(std::ostream &o, const GenTL::growthmodel &g);

#endif
