#ifndef GSL_DISTRIBUTIONS_H__
#define GSL_DISTRIBUTIONS_H__
#include <cmath>
#include <stdexcept>
#include <iostream>
#include <string>

#include "gsl_rand.H"

/** a wrapper class around the gsl distributions                */
/** First a generic distribution                                */
template<typename T,typename PARAMETERS> 
class distribution {
public:
  virtual T sample(rng &r)=0;
  virtual double log_pdf(const T &x) const =0;
  virtual distribution *clone()=0;
  virtual ~distribution(){}
  //  PARAMETERS pars()=0;
protected:
  PARAMETERS _p;
};


class ctsdistribution: 
  public distribution<double,double* > {
public:
  ctsdistribution(int npars) {
    _p = new double(npars);  
  }
  ctsdistribution(double par1, double par2) {
    _p = new double[2];
    _p[0]=par1;
    _p[1]=par2;
  }
  ctsdistribution(double par) {
    _p = new double[1];
    _p[0]=par;
  }
  ctsdistribution() { // for improper distribution
  }
  virtual double sample(rng &r)=0;
  virtual bool proper() const {
    return true;
  }
  ~ctsdistribution() {
    delete[] _p;
  }
  virtual double log_posterior(int x1, int x2) const {
    return 0.0;
  }
  virtual double log_pdf(const double &x) const =0;
};
/** an improper distribution                                             */
class improperuniform: 
  public ctsdistribution {
public:
  improperuniform() 
    :ctsdistribution()  {
  }
  ~improperuniform(){};
  double sample(rng &r) {
    throw std::domain_error("cannot generate from an improper distribtion");
    return 1.0;
  }
  improperuniform *clone() {
    improperuniform *tmp=new improperuniform();
    return tmp;
  }
  virtual bool proper() const  {
    return false;
  }
  double log_pdf(const double &x) const { 
    // note that a specifically written version of this for
    // logarithms may be preferable
    return 0.0;
  }
};
/** Constant values                                                */
class Constant: 
  public ctsdistribution {
public:
  Constant(double val) 
    :ctsdistribution(val)  {
  }
  ~Constant(){};
  double sample(rng &r) {
    return _p[0];
  }
  Constant *clone() {
    Constant *tmp=new Constant(_p[0]);
    return tmp;
  }
  double log_pdf(const double &x) const { 
    // note that a specifically written version of this for
    // logarithms may be preferable
    return 0.0;
  }
};
/** the normal distribution                                       */
class normal: 
  public ctsdistribution {
public:
  normal(double mu, double sigma) 
    :ctsdistribution(mu,sigma)  {
  }
  ~normal(){};
  double sample(rng &r) {
    return _p[0]+r.normal()*_p[1];
  }
  normal *clone() {
    normal *tmp=new normal(mu(),sigma());
    return tmp;
  }
  double mu() const { return _p[0];}
  double sigma() const{ return _p[1];}
  double log_pdf(const double &x) const { 
    // note that a specifically written version of this for
    // logarithms may be preferable
#ifdef USE_R
    return dnorm(x,_p[0],_p[1],1);
#else
    return log(gsl_ran_gaussian_pdf(x-_p[0],_p[1]));
#endif
  }
  bool proper() const {
    return true;
  }
};
/** The Gamma distribution                                      */
class Gamma
  : public ctsdistribution {
public:
  Gamma(double a, double b) 
    :ctsdistribution(a,b) {
  }
  ~Gamma(){};
  double log_pdf(const double &x) const {
#ifdef USE_R
    return dgamma(x,_p[0],_p[1],1);
#else
    return log(gsl_ran_gamma_pdf(x,_p[0],_p[1]));
#endif  
  }
  double sample(rng &r) {
    return r.rgamma(_p[0],_p[1]);
  }
  Gamma *clone() {
    Gamma *tmp=new Gamma(_p[0],_p[1]);
    return tmp;
  }
 bool proper() const {
    return true;
  }
private:
  Gamma();
};

/** The uniform distribution                                      */
class uniform
  : public ctsdistribution {
public:
  uniform(double a, double b) 
    :ctsdistribution(a,b) {
  }
  ~uniform(){};
  double log_pdf(const double &x) const {
    return -log(_p[1]-_p[0]);
  }
  double sample(rng &r) {
    return _p[0]+r()*(_p[1]-_p[0]);
  }
  uniform *clone() {
    uniform *tmp=new uniform(_p[0],_p[1]);
    return tmp;
  }
 bool proper() const {
    return true;
  }
private:
  uniform();
};

/** The Beta distribution                                      */
class Beta
  : public ctsdistribution {
public:
  Beta(double a, double b) 
    :ctsdistribution(a,b) {
  }
  ~Beta(){};
  double log_pdf(const double &x) const {
    return 	 (a()-1.)*log(x)+(b()-1)*log(1.-x)
      -lgamma(a())-lgamma(b())+lgamma(a()+b());
  }
  double sample(rng &r) {
    double x=r.rgamma(a(),1.0);
    double y=r.rgamma(b(),1.0);
    return x/(x+y);
  }
  Beta *clone() {
    Beta *tmp=new Beta(_p[0],_p[1]);
    return tmp;
  }
  bool proper() const {
    return true;
  }
  double a()const {
    return _p[0];
  }
  double b() const {
    return _p[1];
  }
 double log_posterior(int x1, int x2) const {
    return -lgamma(x1+x2+a()+b())
      +lgamma(x2+b()) + lgamma(x1+a()) 
      -lgamma(a())- lgamma(b()) + lgamma(a()+b()) 
      + lgamma(x1+x2+1.)-lgamma(x1+1.) - lgamma(x2+1.);
  }
private:
  Beta();
};

ctsdistribution *ctsdistributionparse(std::istream &is);
ctsdistribution *fromstring(const std::string &s);

template <typename ITOR>
double ldnorm(ITOR start, ITOR lst, double mean, double sd) {
  ITOR curr=start;
  double temp=0.0;
  while (curr!=lst) {
    //  std::cout << *curr-mean <<" " << sd << " " << temp << " " ;
#ifdef USR_R
    temp += pnorm(*(curr),mean,sd,1);
#else
    temp += log(gsl_ran_gaussian_pdf(*(curr)-mean,sd));
#endif
    curr++;
  }
  //  std::cout << std::endl;
  return temp;
}


#endif
