/** @file            Time-stamp: <05/11/30 15:11:01 nijw>  */ 
#ifndef __RNODET__H_IJW
#define __RNODET__H_IJW

#include <list>
#include <vector>
#include <deque>
#include <iostream>
#include <cassert>
#include <sstream>
#include <stdexcept>
#include <algorithm>
#include <iterator>
// utility headers
#include "utilityfunctionals.h"
#include "util.h"
// Genetics Headers
#include "activesites.h"



namespace GenTL {

  const position maxloci=100000000;
  /** forward declarations                             */
  template<class T> class rec_rates;
  template <class T> class rnode;
  template <class T> class recombnode;

  /** a simple class for holding the lines of descent  */
  template<class T> 
    class lines_of_descent: public std::deque<recombnode<T> *> {
    public:
      /** construct from the sample */
      lines_of_descent( std::vector<rnode<T> > &sample)
	: std::deque<recombnode<T> *>(sample.size()) {
	for (size_t i=0;i<sample.size();i++) {
	  this->at(i)=sample[i].first();
	}
      }
      /** add a recombnode         */
      void add(recombnode<T> *a) {  
	push_back(a);
      } 
      /** coalescence -- rearrange the lines-of-descent */
      int coal(rnode<T> *nw,  std::pair<int, int> &wh) {
	this->at(wh.first) = nw->l.front();
	if (this->at(wh.first)->active.empty()) {
	  assert(wh.first>wh.second);
	  this->at(wh.first)=this->back();
	  this->pop_back();
	  this->at(wh.second)=this->back();
	  this->pop_back();
	  return 2;
	}
	this->at(wh.second)= this->back();
	this->pop_back();
	return 1;
      }
      /** print out the current lines                   */
      void print(std::ostream &o) {
	for (size_t i=0;i<this->size();i++) {
	  std::ostringstream s;
	  s <<  i<<": ";
	  this->at(i)->print(o,s.str().c_str());
	}
      }
    private:
      lines_of_descent();
    };
  
  //////////////////////////////////////////////////////////////////////
  /** \brief recombination node
   *
   * A class to build simple recombination trees
   */
  template<class T>
  class recombnode { 
    typedef recombnode* recptr;
  public:
    /// data
    position where;           //< after which position did the recombination occur 
    activetype active;        //< which positions are active 
    double rtime;             //< what was the time of the event that led to this recombnode
    rnode<T> *coal;           //< pointer to the coalescent node from which this derives.
    /// constructor for a leaf
    recombnode(int loci, rnode<T> *here)
      : where(maxloci),rtime(0.0),coal(here) {
      active.insertrange(0,loci);
    }
    // constructor after a recombination
    recombnode(recptr old, const int wh, double ti)
    : where(wh),rtime(ti),coal(old->coal) {
      assert(where>=old->first());
      assert(where< old->last());    
      active.split(wh,old->active);
    }
    /** constructor after a coalescence */
    recombnode(const recptr l,const recptr r, const double ti
	       ,rec_rates<T> &rr, std::vector<rnode<T> *> &root
	       , rnode<T> *here):where(maxloci), rtime(ti),coal(here)  {
      std::list<int> rts = rr.remove(active.combine(l->active,r->active));
      if (rts.size()>0) { 
	std::list<int>::iterator i=rts.begin();
	while (i!=rts.end()) {
	  root[*i]=here;
	  active.erase(*i);
	  i++;
	}
      }
    }
    /** the first active position                                             */
    position first() { 
      return active.first();
    }
    /** which positions are present in a recombnode?                          */
    std::list<position> present() {
      std::list<position> lp;
      active.present(lp);
      return lp;
    }
    /** the last active postion                                               */
    position last() {
	assert(!active.empty());
	return active.last();
    } 
    /** print the active sites at a recombnode                                */
    std::ostream &print(std::ostream &o, const char *message) {
      o << message;
      active.print(o);
      return o;
    };
    /** Is the position loc "active" at a particular recombnode               */ 
    bool IsActive(position loc) {
      return active.present(loc);
    }
    /** Is this recombnode a leaf node ?                                      */
    bool isleaf() const  {
      if (coal->left_==0) return true;
      return false;
    }
    /** findnode - find a node that is active at pos 
     * find a node that is either an internal node at pos or a leaf 
     * below the current recombnode                                           */ 
    rnode<T> *findnode(int pos) {
      recombnode<T> *tmp=this;
      for (;;) {
	assert(IsActive(pos));
	if (tmp->isleaf()) return tmp->coal;

	assert((tmp->coal->left_->IsActive(pos)||tmp->coal->right_->IsActive(pos)));

	if (tmp->coal->left_->IsActive(pos)) {
	  if (tmp->coal->right_->IsActive(pos))
	    return tmp->coal;
	  else tmp=tmp->coal->left_;
	} else {
	  tmp=tmp->coal->right_;
	}
      } 		
    }
  };
  /////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////
  /** recomblist is a list of recombnodes - another helper class              */
  template<class T>
    class recomblist {
      friend class rnode<T>;

      private:
      std::list<recombnode<T> *> data_;
    public:
      typedef typename std::list<recombnode<T> *>::iterator itor;
      recomblist(){};
      recomblist(recombnode<T> *a) {
	data_.push_back(a);
      }
      itor begin() {
	return data_.begin();
      }
      itor end() {
	return data_.end();
      }
      void splice(itor pos,  recomblist &rlist) {
	//itor iif=find(data_.begin(),data_.end(),val);
	//assert (iif!=data_.end());
	data_.splice(pos,rlist.data_);
      }
      itor erase (itor pos) {
	return data_.erase(pos);
      }
      recombnode<T> *front() {
	return data_.front();
      }
      void push_back(recombnode<T> * x) {
	return data_.push_back(x);
      }
      itor insert(itor pos, recombnode<T>  *x ) {
	return data_.insert(pos,x);
      }
      
    ~recomblist() {
      for_each(data_.begin(),data_.end(),
	       DeleteObject());
    } 
    /** print the active sites in the list of sites        */
    std::ostream &print_active(std::ostream &o, const char *message) {
      if (data_.empty()) {
	o << "no active sites ";
	return o;
      }
      itor i=data_.begin();
      int count=0;
      while (i != data_.end()) {
	o << "address:" << *i << " time:" << (*i)->rtime 
	  << "  recombination " << (*i)->where << " stretch "<< ++count <<  " of "<< data_.size() << ": "; 
	(*i)->print(o,"");
	i++;
      }
      return o;
    };
    /** returns the recombinations before a coalescence      */
    std::set<position> recombpos(){
      std::set<int> r;
      itor i=data_.begin();
      while (i != data_.end()) {
	position ps=(*i)->where;
	r.insert(ps);
	i++; 
      }

      return r;
    }
    void fixlist(recombnode<T> *d) {
      // move all the nodes in the list to the list d->coal 
      // and remove the node pointed to by *d
      // what are the active sites in the descendent node 
      std::list<position> actived=d->active.activesites();
      // now loop through all the segments here
      assert(!data_.empty());
      data_.back()->where=d->where;
      for (itor jj=data_.begin();jj!=data_.end();) {
	// change the coalescent node to d->coal
	(*jj)->coal=d->coal;
	std::list<position> lp = (*jj)->active.activesites();
	std::list<position>::iterator li=lp.begin(),lf;
	while (li!=lp.end()) {
	  lf=find(actived.begin(),actived.end(),*li);
	  if (lf==actived.end()) {
	    (*jj)->active.erase(*li);
	  }
	  li++;
	}
	// erase the empty recombnodes (must be at the front or back)
  	if ((*jj)->active.empty()) { 
 	  delete *jj; 
 	  jj=data_.erase(jj); 
 	} 
 	else jj++; 
      } 
      
     
    
    }
    void fixlist(recombnode<T> *d1, recombnode<T> *d2) {
      std::list<position> actived1=d1->active.activesites();
      std::list<position> actived2=d2->active.activesites();
      // now loop through all the segments here
      for (itor jj=data_.begin();jj!=data_.end();) {
	// change the coalescent node to d->coal
	std::list<position> lp = (*jj)->active.activesites();
	std::list<position>::iterator li=lp.begin(),lf1,lf2;
	while (li!=lp.end()) {
	  lf1=find(actived1.begin(),actived1.end(),*li);
	  lf2=find(actived2.begin(),actived2.end(),*li);
	  if (lf1==actived1.end()&&lf2==actived2.end()) {
	    (*jj)->active.erase(*li);
	  }
	  li++;
	}
	if ((*jj)->active.empty()) {
	  delete *jj;
	  jj=data_.erase(jj);
	} else jj++;
      }
    }	
  };
  ////////////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////

  template <class T>
  class rnode {
    friend class recomblist<T>;
  public:
    /**  public data                                      */
    recombnode<T> *left_,*right_;    //< the recombnodes that are the left and right descendants
    double time;                     //< the time of the coalescence
    recomblist<T> l;                 //< a list of recombnodes that happen through the recombination of the original node
    
    /** a function to return the left and right descendents at position pos 
     * of there is none then throws an exception 
     * it returns 0 if left is a leaf                                            */
    rnode *left(position pos) {
      assert(left_->IsActive(pos));
      return left_->findnode(pos);
    }
    /** right descendent                                     */
    rnode *right(position pos) {
      assert(right_->IsActive(pos)); 
      return right_->findnode(pos);
    };
    /** Constructor after a coalescent event                                            */
    rnode(const std::pair<int, int> &wh, const double t, rec_rates<T> &rr
	  ,std::vector<rnode<T> *> &root, const lines_of_descent<T> &lin)
      :left_(lin[wh.first]),right_(lin[wh.second]),time(t) {
      l.push_back(new recombnode<T>(lin[wh.first],lin[wh.second],t,rr,root,this));
    };
    /** constructor for leaves.          */
      rnode(int loci):left_(0),right_(0),time(0.0) {
      l.push_back(new recombnode<T>(loci,this));
    }; 

    /** return the recombnode after a recombination         */
    recombnode<T> *recombination(recombnode<T> *rn, const position where, const double time) {
	typename recomblist<T>::itor i; 
	for (i=l.begin();i!=l.end();i++) { 
	  if (where < (*i)->last()) break;  	
	} 
	assert(i!=l.end());
	assert(*i==rn);
	// create the new recombination node
	// and insert it into the list before i;
	i=l.insert(i,new recombnode<T>(*i,where,time));   
	return *i;
    }
    /** What is the first recombnode  - overloaded for constant and
     * non-constant recombnodes                                                           */
    recombnode<T> *first() {
      return l.front();
    }
    recombnode<T> *first() const {
      return l.front();
    }
    /** What is the last recombination node - overloaded for constant and
     * non-constant recombnodes                                                           */
    recombnode<T> *last()  const {
      return l.data_.back();
    }
    recombnode<T> *last()  {
      return l.back();
    }
    /** extract the data (but not change it).                                             */
    const T &data() const {
	return data_;
    }; 
    /** extract the data                                                                  */
    T &data() { 
	return data_;
    }; 
    /** which positions are present                                                       */
    std::list<position> present() {
      std::list<position> lp;
      typename recomblist<T>::itor i = l.begin(); 
      while (i!=l.end()) {
	(*i++)->active.present(lp);
      }
      return lp;
    }
    /** what are the min and max active sites here                              */
    std::pair<position,position> minmaxactive() {
      std::pair<position,position> a;
      a.first=l.begin()->active.first();
      a.second=l.back().active.last();
      return a;
    }
    /** print the sites that are active for this line                                     */
    void print_active(std::ostream &o, const char *message) {
	o << message;
	l.print_active(o,"");
    }
    /** check that each of the recombnodes points to the right place                      */
    void check() {
      typename recomblist<T>::itor i=l.begin();
      while (i != l.end()) {
	assert((*i)->coal==this);
	i++;
      }
    }
    /** checkagain */
    bool checknode() {
      typedef std::list<position> lipos;
      lipos a=present();

      if (!isleaf()) {
	lipos::iterator ii=a.begin();
	while(ii!=a.end()) {
	  bool l=left_->IsActive(*ii);
	  bool r=right_->IsActive(*ii);
	  if (!(l||r)) {
	    std::cerr << "problem with node in position " << *ii << std::endl;
	    std::cerr << "left: ";
	    left_->print(std::cerr,"");
	    std::cerr << "right: ";
	    right_->print(std::cerr,"");
	    this->print_active(std::cerr,"");
	    exit(EXIT_FAILURE);
	  }
	  ii++;
	}
      }
      return true;
    }

    /** Is a position active ?                                                            */
    bool IsActive(position loc) {
      typename recomblist<T>::itor i=l.begin();
      while (i != l.end()) {
	if ((*i)->IsActive(loc)) return true;
	i++;
      }
      return false;
    }
    /** Is a recombination position present in this node            */
    bool IsRecombinationActive(position where) {
      if (IsActive(where)&&IsActive(where+1)) return true;
      return false;
    }
    /** length calculation                                                                */
    double length(position pos) {
	double tmp(0.0);
	rnode<T> *l=left(pos);
	if (l->isleaf()) {
	    tmp+=time;
	} else {
	    tmp += time-l->time;
	    tmp += l->length(pos);
	}
	rnode<T> *r=right(pos);
	if (r->isleaf()) { 
	    tmp+=time;
	} else {
	    tmp += time-r->time;
	    tmp += r->length(pos);
	}
	return tmp;
    }
    /**           is this a leaf               */
    bool isleaf() {
      if (left_==0) return true;
      return false;
    }
    /** what are the positions of all recombinationsbefore coalescences  */
    std::set<position> recombinations() {
      return l.recombpos();
    }  

    /** is there any of of the samples that are numbered in wh present in this node                  
     * sample0 is the address of the first sample                         */
    // note that the function below has not been tested
    bool present(const std::vector<size_t> &which, rnode *sample0) {
      if (isleaf()) {
	size_t num=this-sample0;
	std::vector<size_t>::const_iterator it=find(which.begin(),which.end(),num);
	//std::cerr << "--found sample " << num << "----";
	if (it!=which.end()) return true;
	return false;
      }
      bool lft,rght;
      if (left_->coal) 
	lft=left_->coal->present(which,sample0);
      else lft=false;
      if (lft) return true;
      if (right_->coal) rght=right_->coal->present(which,sample0);
      else rght=false;
      return rght;
      //    if (left_->coal->present(which,sample0)) return true;
      //      else return right_->coal->present(which,sample0); 
    }
    /** is there any of of the samples that are numbered in wh present in this node                  
     * sample0 is the address of the first sample                         */
    // note that the function below has not been tested
  
 private:
    T data_;  
    /** Functions that are not defined           */
    rnode(const rnode &a);
  };




}
#endif
