// Time-stamp: <05/11/18 15:13:40 nijw>
/** @file        */
#ifndef RTREE__H_IJW
#define RTREE__H_IJW
#include "gentldeclr.h"
#include "recombinationrates.h"
#include "rnodet.h"

#include "nodet.h"
#include "coalescent.h"
//#include "utilcpp/utilityfunctionals.h" // for DeleteObject
#include "utilityfunctionals.h" // for DeleteObject
#include <list>

namespace GenTL {
  /** The default collector class - just counts the recombinations  *
   * Other collector classes can easily be written - they just have 
   * have the operator()(time, event,which) method which is used within 
   * rtree::simulate                                                    */

  struct recombinations {
    /** default constructor                                             */
    recombinations():rec_(0){};
    /** function overload operator - needed for any COLLECTOR structure */
    void operator()(double time, EventType e,int which=0) {
      if (e==Recombination) rec_ +=1;
    };
    /** function operator that just returns the result                  */
    long operator()() {
      return result();
    };
    /** Return the number of recombinations                             */
    long result() {
      return rec_;
    };
    /** Data -- keeps the number of recombinations                      */
    long rec_;
  };
  struct recombinationLocations {
    /** default constructor                                             */
    recombinationLocations() {};
    /** function overload operator - needed for any COLLECTOR structure */
    void operator()(double time, EventType e,int which=0) {
      if (e==Recombination) where_.push_back(which);
    };
    /** function operator that just returns the result                  */
    std::list<int> &operator()() {
      return where_;
    };
    /** Return the list of recombination locations                      */
    std::list<int> &result() {
      return where_;
    };
    /** Data -- keeps the number of recombinations                      */
    std::list<int> where_;
  };
  /** rtree class.  The genral class for simulating recombination trees */
  template<class T,typename COLLECTOR> 
  class rtree {
  public:
    /** constructor which creates its own coalescent object             */
    rtree(int ss, int nl,const std::vector<double> &recomb_rates,growthmodel *g,rng &r,COLLECTOR &clt)
      :sample(ss,nl)
      ,root(nl)
      ,collect(clt) {
#ifdef CHECK
      std::cerr << "now for the coalescent" << std::endl;
#endif
      coalescent cc(ss,g,r);
      simulate(cc,recomb_rates,r);
    };
    /** Constructor which requires a coalescent object (which includes the sample size          */
      rtree(int nl,const std::vector<double> &recomb_rates, coalescent &c,rng &r,COLLECTOR &clt)
      :root(nl), sample(c.nleft(),nl) ,collect(clt) {
#ifdef CHECK
      std::cerr << "already have the coalescent" << std::endl;
#endif
      simulate(c,recomb_rates,r);
    };
     /** Construct a tree from another tree and the numbers of samples to take   
	rtree(const rtree &otree, std::vector<size_t> &wh) {
	  std::list<rnode<T>  *>::const_iterator i=otree.ancestors.begin();
	  while (i != otree.ancestors.end()) {
	    if (i->left().present(wh,&otree.sample[0])) {
	      
	      
	    }
	  }
	} 
	*/
    /** dtor                 
     * This particular formulation is that of Meyers - effective STL                             */
    ~rtree() {
      for_each(ancestors.begin(),ancestors.end(),
	       DeleteObject());
    }
    /** What is the time of the root at position loc                                             */
    double roottime(position loc) {
      return root[loc]->time;
    };
    /** What is the length of the tree at position loc                                           */
    double length(position loc) {
      return root[loc]->length(loc);
    } 
    /** How many coalescent ndoes are there in this tree                                         */
    long CoalescentNodes() const { 
      return ancestors.size();
    };
    /** randomise the node ordering                                                              */
    void randomise(rng &ran) {
      typename std::list<rnode<T> *>::iterator i=ancestors.begin();
      while (i != ancestors.end()) {
	if (ran()<0.5) {
	  recombnode<T> *tmp=(*i)->left_;
	  (*i)->left_=(*i)->right_;
	  (*i)->right_=tmp;
	} 
	i++;
      }
    }
    /** Simulate the tree                                                                       */
    void simulate(coalescent &c, const std::vector<double> recomb_rates,rng &r);
    std::vector<rnode<T>* > root;
    std::vector<rnode<T> > sample;
    std::list<rnode<T>* > ancestors;
    /** How many loci?                                                                          */
    int nloc() const  {
      return sample[0].last()->active.last()+1;
    }
  private:
    COLLECTOR &collect;
    rtree();
  };

  /** A utility function to do the simulations for all constructors                             */
  template <class T,typename COLLECTOR>
  void rtree<T,COLLECTOR>::simulate(coalescent &c, const std::vector<double> recomb_rates,rng &r)
  {
#ifdef CHECK
    std::cerr << "Before simulation have " << c.nleft() << " lines" << std::endl;
#endif
    double t=0.0;
    lines_of_descent<T> lines(sample);
    rec_rates<T> rr(c.nleft(),recomb_rates.size()+1,recomb_rates,r);
    for (;;) {
#ifdef CHECK
      std::cerr << "================================================" << std::endl 
		<< "Time is now " << t << std::endl;
      lines.print(std::cout);
      std::cerr << "total rate "<<rr.rate()<<std::endl;
      rr.print_rates(std::cerr);
      rr.check(lines);
      c.print(std::cerr);
#endif    
      double NextRecombination=rr.next(t);
      double NextEvent=c.next(t);
#ifdef CHECK
      std::cerr << "NextRecombination = " << NextRecombination 
		<< "Next Event " << NextEvent << std::endl;
#endif
      if (NextEvent<NextRecombination) {
	t += NextEvent;
	std::pair<int,int> chromo = c.coalchromo();
#ifdef CHECK
	std::cout << "coalescing "<< chromo.first 
		  << " and " << chromo.second << std::endl;
#endif
	if (chromo.first!=chromo.second) { // a coalescence 
	  collect(t,Coalescence);
	  ancestors.push_back(new rnode<T>(chromo,t,rr,root,lines));
	  // Now fix lines 
	  int removed=lines.coal(ancestors.back(),chromo);
	  c.UpdateCoalescence(chromo,removed);    
	  rr.update_coal(chromo,lines[chromo.first]->active,removed); 
	} else collect(t,Migration);  // other
      } else {//	  recombination
	t += NextRecombination;
	std::pair <int,int> wh=rr.recombination(lines);
	collect(t,Recombination,wh.second);
#ifdef CHECK
	std::cout << "recombination for line " << wh.first
		  << " at position " << wh.second << std::endl;
#endif
	recombnode <T> *nw=
	  lines[wh.first]->coal->recombination(lines[wh.first],wh.second,t);	    
      
	lines.push_back(nw);
	c.UpdateRecombination(wh.first);
	rr.update_recombination(wh.first,lines);
      }
      if (c.nleft()<=1) return;
    }
  }
  /** recursive function to extract the underlying tree at a given position */
  template<typename T>
    node<rnode<T> *> *extract_node(rnode<T> *here,int position, rnode <T> *first)
    {
      if (here->isleaf()) 
	return new node<rnode<T> *>(here-first,here);
      else 
	return new  node<rnode<T> *>(here->time
				,extract_node(here->left(position),position,first)
				,extract_node(here->right(position),position,first)
				,here);
    }

  template <typename T>
    void recursively_destroy_node(node <T> *here) {
    if (!here->isleaf()) {
      recursively_destroy_node(here->left);
      recursively_destroy_node(here->right);
    }
    delete here;
  }

 template <typename T>
   int rotate_tree_labels(node <T> *here) {
   if (here->isleaf()) return here->label;
   int lft=rotate_tree_labels(here->left);
   int rgt=rotate_tree_labels(here->right);
   if (lft>rgt) {
     node<T> *tmp=here->left;
     here->left=here->right;
     here->right=tmp;
     return lft;
   }
   return rgt;
 }


}
#endif   

